/*
 * BigMessageSizeQueue
 * This is the queue for put(send), pop(receive), delete message which size is bigger than a given limitation.
 */

import com.aliyun.mns.client.CloudQueue;
import com.aliyun.mns.common.ClientException;
import com.aliyun.mns.common.ServiceException;
import com.aliyun.mns.model.Message;
import com.aliyun.oss.OSS;
import com.aliyun.oss.OSSException;
import com.aliyun.oss.model.OSSObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.*;
import java.util.HashMap;
import java.util.UUID;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class BigMessageSizeQueue {
    private static Logger logger = LoggerFactory.getLogger(BigMessageSizeQueue.class);
    //set default size limitation is 64k.
    private long BIG_MESSAGE_SIZE = 64 * 1024;
    private boolean needDeleteMessageObjectOnOSSFlag = true;
    private String OSS_URL_MESSAGE_PREFIX = "#OSS_MESSAGE_BODY#";

    private CloudQueue innerQueue;
    private OSS ossClient;
    private String ossBucketName;

    private HashMap<String, String> messageId2OssUrlMap;
    private Lock mapLock;

    private String putMessageBodyToOSS(String messageBody) {
        logger.debug("getMessageBodyFromOSS start");
        String objectKey = "big-message-body-" + UUID.randomUUID();
        InputStream instream = new ByteArrayInputStream(messageBody.getBytes());
        this.ossClient.putObject(this.ossBucketName, objectKey, instream);

        String ossUrl = this.ossBucketName + ":" + objectKey;
        logger.debug("putMessageBodyToOSS: ossUrl=" + ossUrl);
        return ossUrl;
    }

    private String getMessageBodyFromOSS(String ossUrl) {
        logger.debug("getMessageBodyFromOSS ossUrl=" + ossUrl);
        String[] info = ossUrl.split(":");
        String bucketName = info[0];
        String objectKey = info[1];
        logger.debug("getMessageBodyFromOSS: bucketName=" + bucketName +
                " objectKey" + objectKey);

        try {
            OSSObject obj = this.ossClient.getObject(bucketName, objectKey);

            StringBuilder sb = new StringBuilder("");
            InputStream istream = obj.getObjectContent();
            BufferedReader reader = new BufferedReader(new InputStreamReader(istream));
            try {
                while (true) {
                    String line = reader.readLine();
                    if (line == null) {
                        break;
                    }
                    sb.append(line);
                }
            } catch (IOException ioe) {
                logger.error("Exception occurs when getting message body from oss");
                sb = new StringBuilder("Error MessageBody!");
            } finally {
                try {
                    reader.close();
                } catch (IOException ioe) {
                }
            }

            return sb.toString();
        } catch (OSSException e) { // 兼容从mns可能收到重复消息的情况
            if ("NoSuchKey".equals(e.getErrorCode())) {
                return null;
            }
            throw e;
        }
    }

    private void deleteMessageBodyObjectOnOSS(String ossUrl) {
        String[] info = ossUrl.split(":");
        String bucketName = info[0];
        String objectKey = info[1];
        logger.debug("deleteMessageBodyObjectOnOSS: bucketName=" + bucketName
                + " objectKey=" + objectKey);
        this.ossClient.deleteObject(bucketName, objectKey);
    }

    private BigSizeMessage parseMessageBody(Message msg) {
        BigSizeMessage bigSizeMsg = new BigSizeMessage();
        String messageBody = msg.getMessageBodyAsRawString();
        logger.debug("MNS raw message body:" + messageBody);
        if (messageBody.startsWith(this.OSS_URL_MESSAGE_PREFIX)) {
            bigSizeMsg.isBigSizeMessage = true;
            bigSizeMsg.messageBody = messageBody.substring(this.OSS_URL_MESSAGE_PREFIX.length());
        } else {
            bigSizeMsg.isBigSizeMessage = false;
            bigSizeMsg.messageBody = messageBody;
        }
        return bigSizeMsg;
    }

    //public methods start here.
    public BigMessageSizeQueue(CloudQueue cq, OSS ossClient, String ossBucketName) {
        this.innerQueue = cq;
        this.ossClient = ossClient;
        this.ossBucketName = ossBucketName;

        this.messageId2OssUrlMap = new HashMap<>();
        this.mapLock = new ReentrantLock();
    }

    public void setNeedDeleteMessageObjectOnOSSFlag(boolean flag) {
        this.needDeleteMessageObjectOnOSSFlag = flag;
    }

    // message related operations.
    public Message putMessage(Message message) throws ServiceException, ClientException {
        long msgSize = message.getMessageBodyAsBase64().length();
        if (msgSize < BIG_MESSAGE_SIZE) {
            message = this.innerQueue.putMessage(message);
        } else {
            String ossUrl = this.putMessageBodyToOSS(message.getMessageBodyAsRawString());
            String newMessageBody = this.OSS_URL_MESSAGE_PREFIX + ossUrl;
            message.setBaseMessageBody(newMessageBody);
            message = this.innerQueue.putMessage(message);
            logger.info("messageId:{}, ossUrl:{}, messageSize:{}KB, send message by oss", message.getMessageId(), ossUrl, msgSize / 1024);
        }

        return message;
    }

    public Message popMessage(int waitSeconds) {
        Message msg = this.innerQueue.popMessage(waitSeconds);
        if (msg == null) {
            return null;
        }
        BigSizeMessage bsm = this.parseMessageBody(msg);
        if (bsm.isBigSizeMessage) {
            String ossUrl = bsm.messageBody;
            logger.info("messageId:{}, ossUrl:{}, size is over 64KB, receive message from oss", msg.getMessageId(), ossUrl);
            String messageBody = this.getMessageBodyFromOSS(ossUrl);
            if (messageBody == null) {
                return null;
            }
            msg.setMessageBody(messageBody, Message.MessageBodyType.RAW_STRING);

            this.mapLock.lock();
            if (!this.messageId2OssUrlMap.containsKey(msg.getReceiptHandle())) {
                this.messageId2OssUrlMap.put(msg.getReceiptHandle(), ossUrl);
            } else {
                logger.error("ReceiptHandle is already added.");
            }
            this.mapLock.unlock();
        }

        return msg;
    }

    public void deleteMessage(String receiptHandle) {
        this.innerQueue.deleteMessage(receiptHandle);

        //Delete the OSS object if needed. 
        this.mapLock.lock();
        try {
            if (this.messageId2OssUrlMap.containsKey(receiptHandle)) {
                if (this.needDeleteMessageObjectOnOSSFlag) {
                    //Delete the message object on OSS.
                    String ossUrl = this.messageId2OssUrlMap.get(receiptHandle);
                    this.deleteMessageBodyObjectOnOSS(ossUrl);
                    this.messageId2OssUrlMap.remove(receiptHandle);
                }
            }
        } catch (Exception e) {
            logger.error("Exception occurs while deleting message body on oss.");
        } finally {
            this.mapLock.unlock();
        }
    }

    public Message changeMessageVisibility(String receiptHandle,
                                           int visibilityTimeout) throws ServiceException, ClientException {
        return this.innerQueue.changeMessageVisibility(receiptHandle, visibilityTimeout);
    }
}